@@ -15,7 +15,8 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
+ * USA.
  */
 
 #ifndef GNUPG_COMMON_IOBUF_H
@@ -36,6 +37,7 @@
 #define IOBUFCTRL_USER	    16
 
 typedef struct iobuf_struct *iobuf_t;
+typedef struct iobuf_struct *IOBUF;  /* Compatibility with gpg 1.4. */
 
 /* fixme: we should hide most of this stuff */
 struct iobuf_struct
@@ -89,6 +91,7 @@ struct iobuf_struct
 EXTERN_UNLESS_MAIN_MODULE int iobuf_debug_mode;
 
 void iobuf_enable_special_filenames (int yes);
+int  iobuf_is_pipe_filename (const char *fname);
 iobuf_t iobuf_alloc (int use, size_t bufsize);
 iobuf_t iobuf_temp (void);
 iobuf_t iobuf_temp_with_content (const char *buffer, size_t length);
@@ -133,17 +136,18 @@ int iobuf_write_temp (iobuf_t a, iobuf_t temp);
 size_t iobuf_temp_to_buffer (iobuf_t a, byte * buffer, size_t buflen);
 void iobuf_unget_and_close_temp (iobuf_t a, iobuf_t temp);
 
-off_t iobuf_get_filelength (iobuf_t a);
+off_t iobuf_get_filelength (iobuf_t a, int *overflow);
 #define IOBUF_FILELENGTH_LIMIT 0xffffffff
+int  iobuf_get_fd (iobuf_t a);
 const char *iobuf_get_real_fname (iobuf_t a);
 const char *iobuf_get_fname (iobuf_t a);
 
-void iobuf_set_block_mode (iobuf_t a, size_t n);
 void iobuf_set_partial_block_mode (iobuf_t a, size_t len);
-int iobuf_in_block_mode (iobuf_t a);
 
 int iobuf_translate_file_handle (int fd, int for_write);
 
+void iobuf_skip_rest (iobuf_t a, unsigned long n, int partial);
+
 
 /* get a byte form the iobuf; must check for eof prior to this function
  * this function returns values in the range 0 .. 255 or -1 to indicate EOF
