@@ -1,5 +1,6 @@
 /* main.h
- * Copyright (C) 1998, 1999, 2000, 2001, 2002 Free Software Foundation, Inc.
+ * Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
+ *               2006 Free Software Foundation, Inc.
  *
  * This file is part of GnuPG.
  *
@@ -15,28 +16,36 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
+ * USA.
  */
 #ifndef G10_MAIN_H
 #define G10_MAIN_H
+
 #include "types.h"
-#include "gpg.h"
 #include "../common/iobuf.h"
-#include "mpi.h"
 #include "cipher.h"
 #include "keydb.h"
 
 /* It could be argued that the default cipher should be 3DES rather
    than CAST5, and the default compression should be 0
-   (i.e. uncompressed) rather than 1 (zip). */
-#define DEFAULT_CIPHER_ALGO   CIPHER_ALGO_CAST5
-#define DEFAULT_DIGEST_ALGO   DIGEST_ALGO_SHA1
-#define DEFAULT_COMPRESS_ALGO 1
+   (i.e. uncompressed) rather than 1 (zip).  However, the real world
+   issues of speed and size come into play here. */
 
-typedef struct {
-    int header_okay;
-    PK_LIST pk_list;
-    cipher_filter_context_t cfx;
+#define DEFAULT_CIPHER_ALGO     CIPHER_ALGO_CAST5
+#define DEFAULT_DIGEST_ALGO     DIGEST_ALGO_SHA1
+#define DEFAULT_COMPRESS_ALGO   COMPRESS_ALGO_ZIP
+#define DEFAULT_S2K_DIGEST_ALGO DIGEST_ALGO_SHA1
+
+#define S2K_DIGEST_ALGO (opt.s2k_digest_algo?opt.s2k_digest_algo:DEFAULT_S2K_DIGEST_ALGO)
+
+typedef struct
+{
+  int header_okay;
+  PK_LIST pk_list;
+  DEK *symkey_dek;
+  STRING2KEY *symkey_s2k;
+  cipher_filter_context_t cfx;
 } encrypt_filter_context_t;
 
 struct groupitem
@@ -46,7 +55,7 @@ struct groupitem
   struct groupitem *next;
 };
 
-/*-- g10.c --*/
+/*-- gpg.c --*/
 extern int g10_errors_seen;
 
 #if __GNUC__ > 2 || (__GNUC__ == 2 && __GNUC_MINOR__ >= 5 )
@@ -64,18 +73,20 @@ char *make_radix64_string( const byte *data, size_t len );
 /*-- misc.c --*/
 void trap_unaligned(void);
 int disable_core_dumps(void);
+void register_secured_file (const char *fname);
+void unregister_secured_file (const char *fname);
+int  is_secured_file (int fd);
+int  is_secured_filename (const char *fname);
 u16 checksum_u16( unsigned n );
 u16 checksum( byte *p, unsigned n );
 u16 checksum_mpi( gcry_mpi_t a );
 u32 buffer_to_u32( const byte *buffer );
 const byte *get_session_marker( size_t *rlen );
 int openpgp_cipher_test_algo( int algo );
-int openpgp_pk_test_algo( int algo, unsigned int usage_flags );
+int openpgp_pk_test_algo( int algo );
+int openpgp_pk_test_algo2 ( int algo, unsigned int use );
 int openpgp_pk_algo_usage ( int algo );
 int openpgp_md_test_algo( int algo );
-int openpgp_md_map_name (const char *string);
-int openpgp_cipher_map_name (const char *string);
-int openpgp_pk_map_name (const char *string);
 
 #ifdef USE_IDEA
 void idea_cipher_warn( int show );
@@ -91,9 +102,13 @@ struct expando_args
 };
 
 char *pct_expando(const char *string,struct expando_args *args);
-int hextobyte( const char *s );
 void deprecated_warning(const char *configname,unsigned int configlineno,
 			const char *option,const char *repl1,const char *repl2);
+void deprecated_command (const char *name);
+
+int string_to_cipher_algo (const char *string);
+int string_to_digest_algo (const char *string);
+
 const char *compress_algo_to_string(int algo);
 int string_to_compress_algo(const char *string);
 int check_compress_algo(int algo);
@@ -106,10 +121,19 @@ struct parse_options
 {
   char *name;
   unsigned int bit;
+  char **value;
+  char *help;
 };
 
-int parse_options(char *str,unsigned int *options,struct parse_options *opts);
-
+char *optsep(char **stringp);
+char *argsplit(char *string);
+int parse_options(char *str,unsigned int *options,
+		  struct parse_options *opts,int noisy);
+char *unescape_percent_string (const unsigned char *s);
+int has_invalid_email_chars (const char *s);
+int is_valid_mailbox (const char *name);
+const char *get_libexecdir (void);
+int path_access(const char *file,int mode);
 
 /* Temporary helpers. */
 int pubkey_get_npkey( int algo );
@@ -117,31 +141,24 @@ int pubkey_get_nskey( int algo );
 int pubkey_get_nsig( int algo );
 int pubkey_get_nenc( int algo );
 unsigned int pubkey_nbits( int algo, gcry_mpi_t *pkey );
-
-/* MPI helpers. */
-int mpi_write( iobuf_t out, gcry_mpi_t a );
-int mpi_write_opaque( iobuf_t out, gcry_mpi_t a );
-gcry_mpi_t mpi_read(iobuf_t inp, unsigned int *ret_nread, int secure );
-gcry_mpi_t mpi_read_opaque(iobuf_t inp, unsigned int *ret_nread );
 int mpi_print( FILE *fp, gcry_mpi_t a, int mode );
 
 
-
-
 /*-- helptext.c --*/
 void display_online_help( const char *keyword );
 
 /*-- encode.c --*/
+int setup_symkey(STRING2KEY **symkey_s2k,DEK **symkey_dek);
 int encode_symmetric( const char *filename );
 int encode_store( const char *filename );
-int encode_crypt( const char *filename, STRLIST remusr );
+int encode_crypt( const char *filename, STRLIST remusr, int use_symkey );
 void encode_crypt_files(int nfiles, char **files, STRLIST remusr);
 int encrypt_filter( void *opaque, int control,
 		    iobuf_t a, byte *buf, size_t *ret_len);
 
 
 /*-- sign.c --*/
-int complete_sig( PKT_signature *sig, PKT_secret_key *sk, MD_HANDLE md );
+int complete_sig( PKT_signature *sig, PKT_secret_key *sk, gcry_md_hd_t md );
 int sign_file( STRLIST filenames, int detached, STRLIST locusr,
 	       int do_encrypt, STRLIST remusr, const char *outfile );
 int clearsign_file( const char *fname, STRLIST locusr, const char *outfile );
@@ -149,31 +166,43 @@ int sign_symencrypt_file (const char *fname, STRLIST locusr);
 
 /*-- sig-check.c --*/
 int check_revocation_keys (PKT_public_key *pk, PKT_signature *sig);
+int check_backsig(PKT_public_key *main_pk,PKT_public_key *sub_pk,
+		  PKT_signature *backsig);
 int check_key_signature( KBNODE root, KBNODE node, int *is_selfsig );
 int check_key_signature2( KBNODE root, KBNODE node, PKT_public_key *check_pk,
-                          PKT_public_key *ret_pk, int *is_selfsig,
-                          u32 *r_expiredate, int *r_expired );
+			  PKT_public_key *ret_pk, int *is_selfsig,
+			  u32 *r_expiredate, int *r_expired );
 
 /*-- delkey.c --*/
 int delete_keys( STRLIST names, int secret, int allow_both );
 
 /*-- keyedit.c --*/
-void keyedit_menu( const char *username, STRLIST locusr, STRLIST cmds,
-							    int sign_mode );
+void keyedit_menu( const char *username, STRLIST locusr,
+		   STRLIST commands, int quiet, int seckey_check );
 void show_basic_key_info (KBNODE keyblock);
 
 /*-- keygen.c --*/
-u32 ask_expire_interval(int object);
+u32 parse_expire_string(const char *string);
+u32 ask_expire_interval(int object,const char *def_expire);
 u32 ask_expiredate(void);
-void generate_keypair( const char *fname, const char *card_serialno );
+void generate_keypair( const char *fname, const char *card_serialno,
+                       const char *backup_encryption_dir );
 int keygen_set_std_prefs (const char *string,int personal);
 PKT_user_id *keygen_get_std_prefs (void);
 int keygen_add_key_expire( PKT_signature *sig, void *opaque );
 int keygen_add_std_prefs( PKT_signature *sig, void *opaque );
 int keygen_upd_std_prefs( PKT_signature *sig, void *opaque );
 int keygen_add_keyserver_url(PKT_signature *sig, void *opaque);
+int keygen_add_notations(PKT_signature *sig,void *opaque);
 int keygen_add_revkey(PKT_signature *sig, void *opaque);
+int make_backsig(PKT_signature *sig,PKT_public_key *pk,
+		 PKT_public_key *sub_pk,PKT_secret_key *sub_sk);
 int generate_subkeypair( KBNODE pub_keyblock, KBNODE sec_keyblock );
+#ifdef ENABLE_CARD_SUPPORT
+int generate_card_subkeypair (KBNODE pub_keyblock, KBNODE sec_keyblock,
+                              int keyno, const char *serialno);
+int save_unprotected_key_to_card (PKT_secret_key *sk, int keyno);
+#endif
 
 /*-- openfile.c --*/
 int overwrite_filep( const char *fname );
@@ -185,29 +214,29 @@ void try_make_homedir( const char *fname );
 
 /*-- seskey.c --*/
 void make_session_key( DEK *dek );
-gcry_mpi_t encode_session_key( DEK *dek, unsigned int nbits);
-gcry_mpi_t encode_md_value( int pubkey_algo,  MD_HANDLE md,
-		     int hash_algo, unsigned nbits, int v3compathack );
-
-/*-- comment.c --*/
-KBNODE make_comment_node_from_buffer (const char *s, size_t n);
-KBNODE make_comment_node( const char *s );
-KBNODE make_mpi_comment_node( const char *s, gcry_mpi_t a );
+gcry_mpi_t encode_session_key( DEK *dek, unsigned nbits );
+gcry_mpi_t encode_md_value( PKT_public_key *pk, PKT_secret_key *sk,
+                            gcry_md_hd_t md, int hash_algo );
 
 /*-- import.c --*/
-int parse_import_options(char *str,unsigned int *options);
+int parse_import_options(char *str,unsigned int *options,int noisy);
 void import_keys( char **fnames, int nnames,
 		  void *stats_hd, unsigned int options );
-int import_keys_stream( iobuf_t inp,
-			void *stats_hd, unsigned int options );
+int import_keys_stream( iobuf_t inp,void *stats_hd,unsigned char **fpr,
+			size_t *fpr_len,unsigned int options );
 void *import_new_stats_handle (void);
 void import_release_stats_handle (void *p);
 void import_print_stats (void *hd);
 
 int collapse_uids( KBNODE *keyblock );
 
+int auto_create_card_key_stub ( const char *serialnostr, 
+                                const unsigned char *fpr1,
+                                const unsigned char *fpr2,
+                                const unsigned char *fpr3);
+
 /*-- export.c --*/
-int parse_export_options(char *str,unsigned int *options);
+int parse_export_options(char *str,unsigned int *options,int noisy);
 int export_pubkeys( STRLIST users, unsigned int options );
 int export_pubkeys_stream( iobuf_t out, STRLIST users,
 			   KBNODE *keyblock_out, unsigned int options );
@@ -221,7 +250,7 @@ int enarmor_file( const char *fname );
 /*-- revoke.c --*/
 struct revocation_reason_info;
 int gen_revoke( const char *uname );
-int gen_desig_revoke( const char *uname );
+int gen_desig_revoke( const char *uname, STRLIST locusr);
 int revocation_reason_build_cb( PKT_signature *sig, void *opaque );
 struct revocation_reason_info *
 		ask_revocation_reason( int key_rev, int cert_rev, int hint );
@@ -230,17 +259,20 @@ void release_revocation_reason_info( struct revocation_reason_info *reason );
 /*-- keylist.c --*/
 void public_key_list( STRLIST list );
 void secret_key_list( STRLIST list );
+void print_subpackets_colon(PKT_signature *sig);
 void reorder_keyblock (KBNODE keyblock);
 void list_keyblock( KBNODE keyblock, int secret, int fpr, void *opaque );
 void print_fingerprint (PKT_public_key *pk, PKT_secret_key *sk, int mode);
+void print_revokers(PKT_public_key *pk);
 void show_policy_url(PKT_signature *sig,int indent,int mode);
 void show_keyserver_url(PKT_signature *sig,int indent,int mode);
-void show_notation(PKT_signature *sig,int indent,int mode);
+void show_notation(PKT_signature *sig,int indent,int mode,int which);
 void dump_attribs(const PKT_user_id *uid,
 		  PKT_public_key *pk,PKT_secret_key *sk);
 void set_attrib_fd(int fd);
 void print_seckey_info (PKT_secret_key *sk);
 void print_pubkey_info (FILE *fp, PKT_public_key *pk);
+void print_card_key_info (FILE *fp, KBNODE keyblock);
 
 /*-- verify.c --*/
 void print_file_status( int status, const char *name, int what );
@@ -249,19 +281,12 @@ int verify_files( int nfiles, char **files );
 
 /*-- decrypt.c --*/
 int decrypt_message( const char *filename );
-void decrypt_messages(int nfiles, char **files);
+void decrypt_messages(int nfiles, char *files[]);
 
 /*-- plaintext.c --*/
-int hash_datafiles( MD_HANDLE md, MD_HANDLE md2,
+int hash_datafiles( gcry_md_hd_t md, gcry_md_hd_t md2,
 		    STRLIST files, const char *sigfilename, int textmode );
-
-/*-- pipemode.c --*/
-void run_in_pipemode (void);
-
-/*-- card-util.c --*/
-void change_pin (int no, int allow_admin);
-void card_status (FILE *fp, char *serialnobuf, size_t serialnobuflen);
-void card_edit (STRLIST commands);
+PKT_plaintext *setup_plaintext_name(const char *filename,IOBUF iobuf);
 
 /*-- signal.c --*/
 void init_signals(void);
@@ -269,4 +294,14 @@ void pause_on_sigusr( int which );
 void block_all_signals(void);
 void unblock_all_signals(void);
 
+
+#ifdef ENABLE_CARD_SUPPORT
+/*-- card-util.c --*/
+void change_pin (int no, int allow_admin);
+void card_status (FILE *fp, char *serialno, size_t serialnobuflen);
+void card_edit (STRLIST commands);
+int  card_generate_subkey (KBNODE pub_keyblock, KBNODE sec_keyblock);
+int  card_store_subkey (KBNODE node, int use);
+#endif
+
 #endif /*G10_MAIN_H*/
