@@ -15,12 +15,15 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
+ * USA.
  */
 
 #ifndef G10_MEMORY_H
 #define G10_MEMORY_H
 
+#error this file should not be used anymore
+
 #ifdef M_DEBUG
 #ifndef STR
 #define STR(v) #v
@@ -66,7 +69,7 @@ size_t m_size( const void *a );
 void m_print_stats(const char *prefix);
 
 /*-- secmem.c --*/
-int secmem_init( size_t npool );
+void secmem_init( size_t npool );
 void secmem_term( void );
 void *secmem_malloc( size_t size );
 void *secmem_realloc( void *a, size_t newsize );
@@ -90,15 +93,4 @@ unsigned secmem_get_flags(void);
 EXTERN_UNLESS_MAIN_MODULE int memory_debug_mode;
 EXTERN_UNLESS_MAIN_MODULE int memory_stat_debug_mode;
 
-/* To prepare a migration to the xmalloc suite of function as used in
-   1.9 we define a couple of macros. */
-#define xmalloc(n)        m_alloc ((n))
-#define xcalloc(n,m)      m_alloc_clear ((n)*(m))
-#define xmalloc_secure(n) m_alloc_secure (n)
-#define xcalloc_secure(n) m_alloc_secure_clear ((n)*(m))
-#define xrealloc(a,n)     m_realloc ((a),(n))
-#define xstrdup(a)        m_strdup ((a))
-#define xfree(a)          m_free (a)
-
-
 #endif /*G10_MEMORY_H*/
