@@ -15,12 +15,13 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
+ * USA.
  */
 #ifndef G10_UTIL_H
 #define G10_UTIL_H
 
-#warning oops, using old util.h
+#error this file should not be used anymore
 #if 0 /* Dont use it anymore */
 
 #if defined (__MINGW32__) || defined (__CYGWIN32__)
