@@ -15,7 +15,8 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
+ * USA.
  */
 
 #include <config.h>
@@ -47,7 +48,7 @@
 #define set_error(e,t) assuan_set_error (ctx, ASSUAN_ ## e, (t))
 
 
-/* Macro to flag a a removed card.  */
+/* Macro to flag a removed card.  */
 #define TEST_CARD_REMOVAL(c,r)                              \
        do {                                                 \
           int _r = (r);                                     \
@@ -62,9 +63,26 @@
       && (c)->reader_slot == locked_session->ctrl_backlink->reader_slot)
 
 
+/* This structure is used to keep track of open readers (slots). */
+struct slot_status_s 
+{
+  int valid;  /* True if the other objects are valid. */
+  int slot;   /* Slot number of the reader or -1 if not open. */
+
+  int reset_failed; /* A reset failed. */
+
+  int any;    /* Flag indicating whether any status check has been
+                 done.  This is set once to indicate that the status
+                 tracking for the slot has been initialized.  */
+  unsigned int status;  /* Last status of the slot. */
+  unsigned int changed; /* Last change counter of teh slot. */
+};
+
+
 /* Data used to associate an Assuan context with local server data.
    This object describes the local properties of one session.  */
-struct server_local_s {
+struct server_local_s 
+{
   /* We keep a list of all active sessions with the anchor at
      SESSION_LIST (see below).  This field is used for linking. */
   struct server_local_s *next_session; 
@@ -86,6 +104,10 @@ struct server_local_s {
 };
 
 
+/* The table with information on all used slots. */
+static struct slot_status_s slot_table[10];
+
+
 /* To keep track of all running sessions, we link all active server
    contexts and the anchor in this variable.  */
 static struct server_local_s *session_list;
@@ -94,6 +116,13 @@ static struct server_local_s *session_list;
    in this variable. */
 static struct server_local_s *locked_session;
 
+/* While doing a reset we need to make sure that the ticker does not
+   call scd_update_reader_status_file while we are using it. */
+static pth_mutex_t status_file_update_lock = PTH_MUTEX_INIT;
+
+
+/*-- Local prototypes --*/
+static void update_reader_status_file (void);
 
 
 
@@ -107,7 +136,11 @@ update_card_removed (int slot, int value)
   for (sl=session_list; sl; sl = sl->next_session)
     if (sl->ctrl_backlink
         && sl->ctrl_backlink->reader_slot == slot)
-      sl->card_removed = value;
+      {
+        sl->card_removed = value;
+      }
+  if (value)
+    application_notify_card_removed (slot);
 }
 
 
@@ -124,69 +157,84 @@ has_option (const char *line, const char *name)
 }
 
 
-/* Reset the card and free the application context.  With DO_CLOSE set
-   to true and this is the last session with a reference to the
-   reader, close the reader and don't do just a reset. */
+/* Convert the STRING into a newly allocated buffer while translating
+   the hex numbers.  Stops at the first invalid character.  Blanks and
+   colons are allowed to separate the hex digits.  Returns NULL on
+   error or a newly malloced buffer and its length in LENGTH.  */
+static unsigned char *
+hex_to_buffer (const char *string, size_t *r_length)
+{
+  unsigned char *buffer;
+  const char *s;
+  size_t n;
+
+  buffer = xtrymalloc (strlen (string)+1);
+  if (!buffer)
+    return NULL;
+  for (s=string, n=0; *s; s++)
+    {
+      if (spacep (s) || *s == ':') 
+        continue;
+      if (hexdigitp (s) && hexdigitp (s+1))
+        {
+          buffer[n++] = xtoi_2 (s);
+          s++;
+        }
+      else
+        break;
+    }
+  *r_length = n;
+  return buffer;
+}
+
+
+
+/* Reset the card and free the application context.  With SEND_RESET
+   set to true actually send a RESET to the reader. */
 static void
-do_reset (ctrl_t ctrl, int do_close)
+do_reset (ctrl_t ctrl, int send_reset)
 {
   int slot = ctrl->reader_slot;
 
+  if (!(slot == -1 || (slot >= 0 && slot < DIM(slot_table))))
+    BUG ();
+
   if (ctrl->app_ctx)
     {
       release_application (ctrl->app_ctx);
       ctrl->app_ctx = NULL;
     }
-  if (ctrl->reader_slot != -1)
+
+  if (slot != -1 && send_reset && !IS_LOCKED (ctrl) )
     {
-      struct server_local_s *sl;
-
-      /* If we are the only session with the reader open we may close
-         it.  If not, do a reset unless the a lock is held on the
-         reader.  */
-      for (sl=session_list; sl; sl = sl->next_session)
-        if (sl != ctrl->server_local
-            && sl->ctrl_backlink->reader_slot == ctrl->reader_slot)
-          break;
-      if (sl) /* There is another session with the reader open. */
+      if (apdu_reset (slot)) 
         {
-          if ( IS_LOCKED (ctrl) ) /* If it is locked, release it. */
-            ctrl->reader_slot = -1;
-          else
-            {
-              if (do_close) /* Always mark reader unused. */
-                ctrl->reader_slot = -1;
-              else if (apdu_reset (ctrl->reader_slot)) /* Reset only if
-                                                          not locked */
-                {
-                  /* The reset failed.  Mark the reader as closed. */
-                  ctrl->reader_slot = -1;
-                }
-
-              if (locked_session && ctrl->server_local == locked_session)
-                {
-                  locked_session = NULL;
-                  log_debug ("implicitly unlocking due to RESET\n");
-                }
-            }
-        }
-      else /* No other session has the reader open.  */
-        {
-          if (do_close || apdu_reset (ctrl->reader_slot))
-            {
-              apdu_close_reader (ctrl->reader_slot);
-              ctrl->reader_slot = -1;
-            }
-          if ( IS_LOCKED (ctrl) )
-            {
-              log_debug ("WARNING: cleaning up stale session lock\n");
-              locked_session =  NULL;
-            }
+          slot_table[slot].reset_failed = 1;
         }
     }
+  ctrl->reader_slot = -1;
 
-  /* Reset card removed flag for the current reader.  */
+  /* If we hold a lock, unlock now. */
+  if (locked_session && ctrl->server_local == locked_session)
+    {
+      locked_session = NULL;
+      log_info ("implicitly unlocking due to RESET\n");
+    }
+
+  /* Reset card removed flag for the current reader.  We need to take
+     the lock here so that the ticker thread won't concurrently try to
+     update the file.  Note that the update function will set the card
+     removed flag and we will later reset it - not a particualar nice
+     way of implementing it but it works. */
+  if (!pth_mutex_acquire (&status_file_update_lock, 0, NULL))
+    {
+      log_error ("failed to acquire status_fle_update lock\n");
+      return;
+    }
+  update_reader_status_file ();
   update_card_removed (slot, 0);
+  if (!pth_mutex_release (&status_file_update_lock))
+    log_error ("failed to release status_file_update lock\n");
 }
 
 
@@ -195,7 +243,7 @@ reset_notify (assuan_context_t ctx)
 {
   ctrl_t ctrl = assuan_get_pointer (ctx); 
 
-  do_reset (ctrl, 0);
+  do_reset (ctrl, 1);
 }
 
 
@@ -224,18 +272,22 @@ option_handler (assuan_context_t ctx, const char *key, const char *value)
 static int
 get_reader_slot (void)
 {
-  struct server_local_s *sl;
-  int slot= -1;
+  struct slot_status_s *ss;
 
-  for (sl=session_list; sl; sl = sl->next_session)
-    if (sl->ctrl_backlink
-        && (slot = sl->ctrl_backlink->reader_slot) != -1)
-      break;
+  ss = &slot_table[0]; /* One reader for now. */
 
-  if (slot == -1)
-    slot = apdu_open_reader (opt.reader_port);
+  /* Initialize the item if needed. */
+  if (!ss->valid)
+    {
+      ss->slot = -1;
+      ss->valid = 1;
+    }
 
-  return slot;
+  /* Try to open the reader. */
+  if (ss->slot == -1)
+    ss->slot = apdu_open_reader (opt.reader_port);
+
+  return ss->slot;
 }
 
 /* If the card has not yet been opened, do it.  Note that this
@@ -257,7 +309,12 @@ open_card (ctrl_t ctrl, const char *apptype)
     return gpg_error (GPG_ERR_LOCKED);
 
   if (ctrl->app_ctx)
-    return 0; /* Already initialized for one specific application. */
+    {
+      /* Already initialized for one specific application.  Need to
+         check that the client didn't requested a specific application
+         different from the one in use. */
+      return check_application_conflict (ctrl, apptype);
+    }
 
   if (ctrl->reader_slot != -1)
     slot = ctrl->reader_slot;
@@ -342,7 +399,7 @@ cmd_serialno (assuan_context_t ctx, char *line)
     {
       if ( IS_LOCKED (ctrl) )
         return gpg_error (GPG_ERR_LOCKED);
-      do_reset (ctrl, 0);
+      do_reset (ctrl, 1);
     }
 
   if ((rc = open_card (ctrl, *line? line:NULL)))
@@ -684,7 +741,9 @@ pin_cb (void *opaque, const char *info, char **retstr)
 }
 
 
-/* PKSIGN <hexified_id>
+/* PKSIGN [--hash=[rmd160|sha1|md5]] <hexified_id>
+
+   The --hash option is optional; the default is SHA1.
 
  */
 static int
@@ -695,6 +754,26 @@ cmd_pksign (assuan_context_t ctx, char *line)
   unsigned char *outdata;
   size_t outdatalen;
   char *keyidstr;
+  int hash_algo;
+
+  if (has_option (line, "--hash=rmd160"))
+    hash_algo = GCRY_MD_RMD160;
+  else if (has_option (line, "--hash=sha1"))
+    hash_algo = GCRY_MD_SHA1;
+  else if (has_option (line, "--hash=md5"))
+    hash_algo = GCRY_MD_MD5;
+  else if (!strstr (line, "--"))
+    hash_algo = GCRY_MD_SHA1; 
+  else
+    return set_error (Parameter_Error, "invalid hash algorithm");
+  /* Skip over options. */
+  while ( *line == '-' && line[1] == '-' )
+    {
+      while (*line && !spacep (line))
+        line++;
+      while (spacep (line))
+        line++;
+    }
 
   if ( IS_LOCKED (ctrl) )
     return gpg_error (GPG_ERR_LOCKED);
@@ -710,7 +789,7 @@ cmd_pksign (assuan_context_t ctx, char *line)
     return ASSUAN_Out_Of_Core;
   
   rc = app_sign (ctrl->app_ctx,
-                 keyidstr, GCRY_MD_SHA1,
+                 keyidstr, hash_algo,
                  pin_cb, ctx,
                  ctrl->in_data.value, ctrl->in_data.valuelen,
                  &outdata, &outdatalen);
@@ -753,7 +832,7 @@ cmd_pkauth (assuan_context_t ctx, char *line)
   if (!ctrl->app_ctx)
     return gpg_error (GPG_ERR_UNSUPPORTED_OPERATION);
 
-  /* We have to use a copy of the key ID because the function may use
+ /* We have to use a copy of the key ID because the function may use
      the pin_cb which in turn uses the assuan line buffer and thus
      overwriting the original line with the keyid */
   keyidstr = xtrystrdup (line);
@@ -1201,7 +1280,7 @@ cmd_checkpin (assuan_context_t ctx, char *line)
 
    Grant exclusive card access to this session.  Note that there is
    no lock counter used and a second lock from the same session will
-   get ignore.  A single unlock (or RESET) unlocks the session.
+   be ignored.  A single unlock (or RESET) unlocks the session.
    Return GPG_ERR_LOCKED if another session has locked the reader.
 
    If the option --wait is given the command will wait until a
@@ -1225,9 +1304,12 @@ cmd_lock (assuan_context_t ctx, char *line)
 #ifdef USE_GNU_PTH
   if (rc && has_option (line, "--wait"))
     {
+      rc = 0;
       pth_sleep (1); /* Better implement an event mechanism. However,
                         for card operations this should be
                         sufficient. */
+      /* FIXME: Need to check that the connection is still alive.
+         This can be done by issuing status messages. */
       goto retry;
     }
 #endif /*USE_GNU_PTH*/
@@ -1293,6 +1375,131 @@ cmd_getinfo (assuan_context_t ctx, char *line)
 }
 
 
+/* RESTART
+
+   Restart the current connection; this is a kind of warm reset.  It
+   deletes the context used by this connection but does not send a
+   RESET to the card.  Thus the card itself won't get reset. 
+
+   This is used by gpg-agent to reuse a primary pipe connection and
+   may be used by clients to backup from a conflict in the serial
+   command; i.e. to select another application. 
+*/
+
+static int
+cmd_restart (assuan_context_t ctx, char *line)
+{
+  ctrl_t ctrl = assuan_get_pointer (ctx);
+
+  if (ctrl->app_ctx)
+    {
+      release_application (ctrl->app_ctx);
+      ctrl->app_ctx = NULL;
+    }
+  if (locked_session && ctrl->server_local == locked_session)
+    {
+      locked_session = NULL;
+      log_info ("implicitly unlocking due to RESTART\n");
+    }
+  return 0;
+}
+
+
+/* APDU [--atr] [--more] [hexstring]
+
+   Send an APDU to the current reader.  This command bypasses the high
+   level functions and sends the data directly to the card.  HEXSTRING
+   is expected to be a proper APDU.  If HEXSTRING is not given no
+   commands are set to the card but the command will implictly check
+   whether the card is ready for use. 
+
+   Using the option "--atr" returns the ATR of the card as a status
+   message before any data like this:
+     S CARD-ATR 3BFA1300FF813180450031C173C00100009000B1
+
+   Using the option --more handles the card status word MORE_DATA
+   (61xx) and concatenate all reponses to one block.
+
+ */
+static int
+cmd_apdu (assuan_context_t ctx, char *line)
+{
+  ctrl_t ctrl = assuan_get_pointer (ctx);
+  int rc;
+  int rc_is_assuan = 0;
+  unsigned char *apdu;
+  size_t apdulen;
+  int with_atr;
+  int handle_more;
+
+  with_atr = has_option (line, "--atr");
+  handle_more = has_option (line, "--more");
+
+  /* Skip over options. */
+  while ( *line == '-' && line[1] == '-' )
+    {
+      while (*line && !spacep (line))
+        line++;
+      while (spacep (line))
+        line++;
+    }
+
+  if ( IS_LOCKED (ctrl) )
+    return gpg_error (GPG_ERR_LOCKED);
+
+  if ((rc = open_card (ctrl, NULL)))
+    return rc;
+
+  if (with_atr)
+    {
+      unsigned char *atr;
+      size_t atrlen;
+      int i;
+      char hexbuf[400];
+      
+      atr = apdu_get_atr (ctrl->reader_slot, &atrlen);
+      if (!atr || atrlen > sizeof hexbuf - 2 )
+        {
+          rc = gpg_error (GPG_ERR_INV_CARD);
+          goto leave;
+        }
+      for (i=0; i < atrlen; i++)
+        sprintf (hexbuf+2*i, "%02X", atr[i]);
+      xfree (atr);
+      send_status_info (ctrl, "CARD-ATR", hexbuf, strlen (hexbuf), NULL, 0);
+    }
+
+  apdu = hex_to_buffer (line, &apdulen);
+  if (!apdu)
+    {
+      rc = gpg_error_from_errno (errno);
+      goto leave;
+    }
+  if (apdulen)
+    {
+      unsigned char *result = NULL;
+      size_t resultlen;
+
+      rc = apdu_send_direct (ctrl->reader_slot, apdu, apdulen, handle_more,
+                             &result, &resultlen);
+      if (rc)
+        log_error ("apdu_send_direct failed: %s\n", gpg_strerror (rc));
+      else
+        {
+          rc_is_assuan = 1;
+          rc = assuan_send_data (ctx, result, resultlen);
+          xfree (result);
+        }
+    }
+  xfree (apdu);
+
+ leave:
+  TEST_CARD_REMOVAL (ctrl, rc);
+  return rc_is_assuan? rc : map_to_assuan_status (rc);
+}
+
+
+
 
 
 /* Tell the assuan library about our commands */
@@ -1323,6 +1530,8 @@ register_commands (assuan_context_t ctx)
     { "LOCK",         cmd_lock },
     { "UNLOCK",       cmd_unlock },
     { "GETINFO",      cmd_getinfo },
+    { "RESTART",      cmd_restart },
+    { "APDU",         cmd_apdu },
     { NULL }
   };
   int i, rc;
@@ -1421,7 +1630,7 @@ scd_command_handler (int fd)
     }
 
   /* Cleanup.  */
-  do_reset (&ctrl, 1); 
+  do_reset (&ctrl, 0); 
 
   /* Release the server object.  */
   if (session_list == ctrl.server_local)
@@ -1491,77 +1700,88 @@ send_status_info (ctrl_t ctrl, const char *keyword, ...)
 }
 
 
-/* This function is called by the ticker thread to check for changes
-   of the reader stati.  It updates the reader status files and if
-   requested by the caller also send a signal to the caller.  */
-void
-scd_update_reader_status_file (void)
+/* This is the core of scd_update_reader_status_file but the caller
+   needs to take care of the locking. */
+static void
+update_reader_status_file (void)
 {
-  static struct {
-    int any;
-    unsigned int status;
-    unsigned int changed;
-  } last[10];
-  int slot;
-  int used;
+  int idx;
   unsigned int status, changed;
 
   /* Note, that we only try to get the status, because it does not
      make sense to wait here for a operation to complete.  If we are
      busy working with a card, delays in the status file update should
      be acceptable. */
-  for (slot=0; (slot < DIM(last)
-                &&!apdu_enum_reader (slot, &used)); slot++)
-    if (used && !apdu_get_status (slot, 0, &status, &changed))
-      {
-        if (!last[slot].any || last[slot].status != status
-            || last[slot].changed != changed )
-          {
-            char *fname;
-            char templ[50];
-            FILE *fp;
-            struct server_local_s *sl;
+  for (idx=0; idx < DIM(slot_table); idx++)
+    {
+      struct slot_status_s *ss = slot_table + idx;
 
-            log_info ("updating status of slot %d to 0x%04X\n", slot, status);
+      if (!ss->valid || ss->slot == -1)
+        continue; /* Not valid or reader not yet open. */
+      
+      if ( apdu_get_status (ss->slot, 0, &status, &changed) )
+        continue; /* Get status failed. */
+
+      if (!ss->any || ss->status != status || ss->changed != changed )
+        {
+          char *fname;
+          char templ[50];
+          FILE *fp;
+          struct server_local_s *sl;
+
+          log_info ("updating status of slot %d to 0x%04X\n",
+                    ss->slot, status);
             
-            sprintf (templ, "reader_%d.status", slot);
-            fname = make_filename (opt.homedir, templ, NULL );
-            fp = fopen (fname, "w");
-            if (fp)
+          sprintf (templ, "reader_%d.status", ss->slot);
+          fname = make_filename (opt.homedir, templ, NULL );
+          fp = fopen (fname, "w");
+          if (fp)
+            {
+              fprintf (fp, "%s\n",
+                       (status & 1)? "USABLE":
+                       (status & 4)? "ACTIVE":
+                       (status & 2)? "PRESENT": "NOCARD");
+              fclose (fp);
+            }
+          xfree (fname);
+            
+          /* Set the card removed flag for all current sessions.  We
+             will set this on any card change because a reset or
+             SERIALNO request must be done in any case.  */
+          if (ss->any)
+            update_card_removed (ss->slot, 1);
+          
+          ss->any = 1;
+          ss->status = status;
+          ss->changed = changed;
+
+          /* Send a signal to all clients who applied for it.  */
+          for (sl=session_list; sl; sl = sl->next_session)
+            if (sl->event_signal && sl->assuan_ctx)
               {
-                fprintf (fp, "%s\n",
-                         (status & 1)? "USABLE":
-                         (status & 4)? "ACTIVE":
-                         (status & 2)? "PRESENT": "NOCARD");
-                fclose (fp);
-              }
-            xfree (fname);
-
-            /* Set the card removed flag for all current sessions.  We
-               will set this on any card change because a reset or
-               SERIALNO request must be done in any case.  */
-            if (last[slot].any)
-              update_card_removed (slot, 1);
-
-            last[slot].any = 1;
-            last[slot].status = status;
-            last[slot].changed = changed;
-
-
-            /* Send a signal to all clients who applied for it.  */
-            for (sl=session_list; sl; sl = sl->next_session)
-              if (sl->event_signal && sl->assuan_ctx)
-                {
-                  pid_t pid = assuan_get_pid (sl->assuan_ctx);
-                  int signo = sl->event_signal;
-
-                  log_info ("client pid is %d, sending signal %d\n",
-                            pid, signo);
+                pid_t pid = assuan_get_pid (sl->assuan_ctx);
+                int signo = sl->event_signal;
+                
+                log_info ("client pid is %d, sending signal %d\n",
+                          pid, signo);
 #ifndef HAVE_W32_SYSTEM
-                  if (pid != (pid_t)(-1) && pid && signo > 0)
-                    kill (pid, signo);
+                if (pid != (pid_t)(-1) && pid && signo > 0)
+                  kill (pid, signo);
 #endif
-                }
-          }
-      }
+              }
+        }
+    }
+}
+
+/* This function is called by the ticker thread to check for changes
+   of the reader stati.  It updates the reader status files and if
+   requested by the caller also send a signal to the caller.  */
+void
+scd_update_reader_status_file (void)
+{
+  if (!pth_mutex_acquire (&status_file_update_lock, 1, NULL))
+    return; /* locked - give up. */
+  update_reader_status_file ();
+  if (!pth_mutex_release (&status_file_update_lock))
+    log_error ("failed to release status_file_update lock\n");
 }
